use crate::prelude::*;
use crate::{
    core::RemoteUrl,
    git::{self, repository::open::OpenRepositoryLike},
    gitlib::Direction,
};

#[derive(Debug, thiserror::Error)]
pub enum GitRemoteValidationError {
    #[error("No default push remote configured in repository")]
    NoDefaultPushRemote,
    #[error("No default fetch remote configured in repository")]
    NoDefaultFetchRemote,
    #[error("Push remote mismatch: found {found}, expected {expected}")]
    PushRemoteMismatch {
        found: RemoteUrl,
        expected: RemoteUrl,
    },
    #[error("Fetch remote mismatch: found {found}, expected {expected}")]
    FetchRemoteMismatch {
        found: RemoteUrl,
        expected: RemoteUrl,
    },
    #[error("IO error accessing repository config")]
    Io(#[from] std::io::Error),
    #[error(transparent)]
    Other(#[from] crate::Error),
}

pub type GitRemoteValidationResult<T> = std::result::Result<T, GitRemoteValidationError>;

#[instrument(skip_all)]
pub fn validate_default_remotes(
    open_repository: &dyn OpenRepositoryLike,
    repo_details: &git::RepoDetails,
) -> GitRemoteValidationResult<()> {
    let fetch_remote = open_repository
        .find_default_remote(Direction::Fetch)
        .map_err(|_| GitRemoteValidationError::NoDefaultFetchRemote)?;

    let push_remote = open_repository
        .find_default_remote(Direction::Push)
        .unwrap_or_else(|_| fetch_remote.clone());

    let expected_remote_url = repo_details.remote_url().ok_or_else(|| {
        GitRemoteValidationError::Other(err!("Unable to build forge URL from repo details"))
    })?;

    if !expected_remote_url.matches(&push_remote) {
        return Err(GitRemoteValidationError::PushRemoteMismatch {
            found: push_remote,
            expected: expected_remote_url,
        });
    }

    if !expected_remote_url.matches(&fetch_remote) {
        return Err(GitRemoteValidationError::FetchRemoteMismatch {
            found: fetch_remote,
            expected: expected_remote_url,
        });
    }

    Ok(())
}
